#include "mbed.h"


/* Partie bluetooth*/
#include "MOD24_NRF.h"
#define TRANSFER_SIZE  3
// Canal de communication avec Teraterm
Serial          debug_pc(USBTX, USBRX);
// 
InterruptIn     bp_int(USER_BUTTON);
char    dataToSend[TRANSFER_SIZE] = {12, 9, 5};
char    dataReceived[TRANSFER_SIZE] = {0};
nRF24L01P       nRF24_mod(PA_7, PA_6, PA_5, PA_12, PA_11, PB_12);
// MOSI, MISO, SCK, CSN, CE, IRQ
char    rxDataCnt;


/* Partie Mesure*/
#include "TEMPHUM_14_CLICK.h"
DigitalOut led1(LED1);
/// Create an I2C interface with specific pins : SDA / SCL
I2C                 my_i2c(A4, A5); // SDA / SCL
/// Create a digital output for reset input of the sensor
DigitalOut          my_rst(D9);
/// Create a Color_10_Click module connection with an I2C interface and a Digital Out 
/// for the led
TempHum_14_Click    my_sensor(&my_i2c, &my_rst);
#define WAIT_TIME_MS 500 


// Fonction d'initialisation du module BT nRF24L01
void initNRF24(void);
// Fonction de reception des commandes et d'appelle des fonctions "avancer" et "tourner"
void Reception_NRF24();

// Fonction d'initialisation du protocole i2c pour communiquer avec le capteur 
// TEMP&HUM CLICK 14
void initMesure();
// Fonction d'envoie des mesures de température et d'humidité par bluetooth
void Emission_NRF24(void);

//=================AVANCER===========================
void Demarrer(void);
void Asservissement(void);
void avancer_droite(void);
void avancer_gauche(void);
void Avancer(int distance_commande, double vit);

//=================TOURNER===========================
void tourner_gauche(void);
void tourner_droite(void);
void Tourner(double theta, double vit);



///DEPLACEMENT
int dist;
int angle;
#define N   115             // Nombre d'impulsions pour un tour de roue
#define R   6.00            // Rayon de la roue
#define d   25.5            // Distance entre les deux roues 


// Bornes reliées aux moteurs des roues (les roues sont à l'AVANT)
PwmOut moteur_droite_R(PB_14);      //moteur droit sens reculer
PwmOut moteur_droite_A(PB_13);      //moteur droit sens avancer
PwmOut moteur_gauche_A(PB_8);       //moteur gauche sens avancer
PwmOut moteur_gauche_R(PC_9);       //moteur gauche sens reculer

// Fonctions d'interruption permettant de détecter la rotation des roues
InterruptIn codeur_gauche(PB_4);
InterruptIn codeur_droite(PB_2);

// Variables d'asservissement en direction
double erreur = 0;          //ecart angulaire entre les deux roues
double Ierreur = 0;         //Integrale de l'erreur
double Kp = 0.05;           //Gain correcteur proportionnel
double Ki = 0.00001;        //Gain correcteur intégral
double rc;                  // Rapport cyclique

// Variables de consigne (distance ou rotation)
double  distance_c;         //consigne de distance en cm
double  distance_p = 0;     //distance parcourue
int     sens;               //1 si le robot va vers l'avant
double  theta_c;            //consigne de rotation en degrés

// Variables de position angulaire de chaque roue
double angle_gauche=0;      //roue gauche
double angle_droite=0;      //roue droite
double fcorr;               //facteur de correction pour l'inertie

// Résolution
// Distance parcourue pour 1 impulsion
double delta_d = (2*3.14159265358979323*R)/(1.0*N);    
// Angle réel du robot pour 1 impulsion
double delta = (360*R)/(d*N);               


/* Main Function */
int main() {
    /* Partie bluetooth*/
    initNRF24();
    /* Partie Mesure*/
    initMesure();
    
    while(1) {
        Emission_NRF24();
        Reception_NRF24();
    }
}


// Fonction d'initialisation du module BT nRF24L01
void initNRF24()
{
    nRF24_mod.powerUp();
    wait_us(100000);
    nRF24_mod.setAirDataRate(NRF24L01P_DATARATE_250_KBPS);
    nRF24_mod.setRfFrequency(2400);
    wait_us(100000);
    debug_pc.printf( "nRF24L01+ Frequency    : %d MHz\r\n", nRF24_mod.getRfFrequency());
    debug_pc.printf( "nRF24L01+ Output power : %d dBm\r\n", nRF24_mod.getRfOutputPower());
    debug_pc.printf( "nRF24L01+ Data Rate    : %d kbps\r\n", nRF24_mod.getAirDataRate());
    debug_pc.printf( "Transfers are grouped into %d characters\r\n", TRANSFER_SIZE);
    nRF24_mod.setTransferSize( TRANSFER_SIZE );
    nRF24_mod.setReceiveMode();
    nRF24_mod.enable();
}


// Fonction d'initialisation du protocole i2c pour communiquer avec le capteur 
// TEMP&HUM CLICK 14
void initMesure(void){
    int k = 0;
    printf("Test.\n");
    thread_sleep_for(WAIT_TIME_MS);
    /// PowerUp and initialize the module
    my_sensor.resetSensor();
    /// Collect informations about the sensor
    printf("Part ID = %x\r\n", my_sensor.getPartID());
    printf("Diag = %x\r\n", my_sensor.getDiag());
}


//=====================AVANCER=======================
void avancer_gauche(void){
	angle_gauche = angle_gauche + 1;          //Nombre d'impulsions de la roue gauche
    distance_p = distance_p + delta_d;      //Distance parcourue par le robot

    //Condition d'arret des moteurs : on a parcouru la bonne distance
    if(1.0*abs(distance_p)>=0.8*abs(distance_c)){

        //Envoi d'une impulsion pour freiner puis arret
        if (sens==1){
            moteur_gauche_A.write(0);
            moteur_gauche_R.write(1);
            moteur_droite_R.write(1);
            moteur_droite_A.write(0);
        }
        if (sens==0){
            moteur_gauche_A.write(1);
            moteur_gauche_R.write(0);
            moteur_droite_R.write(0);
            moteur_droite_A.write(1);
        }
        
        moteur_gauche_A.write(0);
        moteur_gauche_R.write(0);
        moteur_droite_R.write(0);
        moteur_droite_A.write(0);
        
    }

    else {
        erreur = angle_droite-angle_gauche;     //ecart angulaire entre les roues
        Ierreur += erreur;                      //intégrale de l'erreur

            // On adapte le rapport cyclique des roues (proportionnel intégral)
            if (sens==1){
                moteur_droite_A.write(rc - Kp*erreur - Ki*Ierreur);
                moteur_gauche_A.write(rc + Kp*erreur + Ki*Ierreur);
                moteur_droite_R.write(0);
                moteur_gauche_R.write(0);
            }
            if (sens==0){
                moteur_droite_R.write(rc - Kp*erreur - Ki*Ierreur);
                moteur_gauche_R.write(rc + Kp*erreur + Ki*Ierreur);
                moteur_droite_A.write(0);
                moteur_gauche_A.write(0);
            }
        }
}

void avancer_droite(void){
    angle_droite = angle_droite + 1;    //Nombre d'impulsions de la roue droite
}


void Demarrer(void){
    //Renitialisation
    angle_gauche=0;
    angle_droite=0;
    distance_p=0;

    //Lance le mouvement
    if (sens==1){
        moteur_droite_R.write(0);
        moteur_droite_A.write(rc);
        moteur_gauche_A.write(rc);
        moteur_gauche_R.write(0);
    }
    else{
        moteur_droite_R.write(rc);
        moteur_droite_A.write(0);
        moteur_gauche_A.write(0);
        moteur_gauche_R.write(rc);
    }
    
}


void Avancer(int distance_commande, double vit){

    rc = vit;                           //affectation du rapport cyclique
    distance_c=distance_commande;       //affectation de la distance de consigne

    //Initialisation des moteurs
    moteur_droite_R.period_ms(10);
    moteur_droite_A.period_ms(10);
    moteur_gauche_A.period_ms(10);
    moteur_gauche_R.period_ms(10);

    moteur_droite_R.write(0);           //le robot est à l'arrêt
    moteur_droite_A.write(0);
    moteur_gauche_A.write(0);
    moteur_gauche_R.write(0);

	//Interruptions
    codeur_gauche.rise(&avancer_gauche);
    codeur_gauche.fall(&avancer_gauche);
    codeur_droite.rise(&avancer_droite);
    codeur_droite.fall(&avancer_droite);

    if (distance_c>0){sens=1;}          //affectation du sens
    else{sens=0;}

    // Permet de démarrer le mouvement
    Demarrer();
}

//====================TOURNER================================
void tourner_gauche(void){
	angle_gauche=angle_gauche+delta;

    if(fcorr*abs(theta_c/2)<=1.0*abs(angle_gauche)){
        // On envoie une impulsion pour arreter les moteurs puis on les éteint
        if (sens==1){
            moteur_gauche_A.write(rc);
            moteur_gauche_R.write(0);
        }
        if (sens==0){
            moteur_gauche_A.write(0);
            moteur_gauche_R.write(rc);
        }
        moteur_gauche_A.write(0);
        moteur_gauche_R.write(0);
    }
}

void tourner_droite(void){
	angle_droite=angle_droite+delta;
   if(fcorr*abs(theta_c/2)<=1.0*abs(angle_droite)){
       // On envoie une impulsion pour arreter les moteurs puis on les éteint
        if (sens==1){
            moteur_droite_R.write(rc);
            moteur_droite_A.write(0);
        }
        if (sens==0){
            moteur_droite_R.write(0);
            moteur_droite_A.write(rc);
        }
        moteur_droite_R.write(0);
        moteur_droite_A.write(0);
    }
}

void Tourner(double theta, double vit){
    rc = vit;
    theta_c = theta;
//Initialisation
    moteur_droite_R.period_ms(10);
    moteur_droite_A.period_ms(10);
    moteur_gauche_A.period_ms(10);
    moteur_gauche_R.period_ms(10);

    angle_gauche=0;
    angle_droite=0;

	//Interruptions
    codeur_gauche.rise(&tourner_gauche);
    codeur_gauche.fall(&tourner_gauche);
    codeur_droite.rise(&tourner_droite);
    codeur_droite.fall(&tourner_droite);

	//Détermination du sens de rotation
    if(theta_c>0){
        sens = 0;
    }
    else{sens=1;}

    //Déterminatin du facteur de correction d'inertie (empirique)
    if(abs(theta_c)<=90){fcorr=0.67;}
    if((abs(theta_c)>90)&(abs(theta_c)<180)){fcorr=(0.9-0.67)*(abs(theta_c)-90)/90+0.67;}
    if(abs(theta_c)>180){fcorr=1;}
    //Début de la rotation

    if (sens==1){
        moteur_droite_R.write(0);
        moteur_droite_A.write(rc);
        moteur_gauche_A.write(0);
        moteur_gauche_R.write(rc);
    }
    else{
        moteur_droite_R.write(rc);
        moteur_droite_A.write(0);
        moteur_gauche_A.write(rc);
        moteur_gauche_R.write(0);
    }

}

void Emission_NRF24(void)
{   
	  float TempToSend;
		float HumToSend;

	  /*Lecture des valeurs de température et d'humidité*/
		my_sensor.readTRH(&TempToSend, &HumToSend);
	  printf("T=%f / H=%f \r\n\n ", TempToSend, HumToSend);
	
		/*On suppose que T est compris entre 10°C et 30°C et H entre 30 et 50.*/
	  char T = (TempToSend-10)*255.0/20;
	  char H = (HumToSend-30)*255.0/20;
	  char dataToSend[TRANSFER_SIZE]={T, H, 0};

	  /* Transmission donnée depuis nRF24 */
	  nRF24_mod.setRfFrequency(2400);
	  nRF24_mod.write( NRF24L01P_PIPE_P0, dataToSend, TRANSFER_SIZE );
	  //debug_pc.printf( "SENT\r\n");
		wait_us(100000);
}

void Reception_NRF24(void)
{   
    int A;  //Valeur d'angle de rotation
    int D;  //Valeur de distance à parcourir
    //int S;  //Sens soit de rotation, soit de translation
    char dataReceived[TRANSFER_SIZE]={0};
    /* Lecture donnée depuis nRF24 */
    if (nRF24_mod.readable())
    {
        // ...read the data into the receive buffer
        rxDataCnt = nRF24_mod.read(NRF24L01P_PIPE_P0, dataReceived, TRANSFER_SIZE);

        if (dataReceived[0]==0){  //0 correspond à la translation et 1 à la rotation
            //Récupère le sens de translation
            if (dataReceived[1]==1){  //1 correspond à avancer et 0 à reculer
                D=dataReceived[2];  //Récupère la valeur de translation avant
            }
            if (dataReceived[1]==0){
                D=-dataReceived[2];  //Récupère la valeur de translation arrière
            }
        }

				//Récupère le sens de rotation
        if (dataReceived[0]==1){
            if (dataReceived[1]==1){  //1 correspond à la gauche et 0 à la droite
                A=dataReceived[2];  //Récupère la valeur de rotation gauche
            }
            if (dataReceived[1]==0){
                A=-dataReceived[2];  //Récupère la valeur de rotation droit
            }
        }
				//On vérifie si le robot a reçu une nouvelle commande de distance
        if ((D!=0)&(D!=dist)){
            dist=D;
            Avancer(D, 0.4);
        }
				//On vérifie si le robot a reçu une nouvelle commande d'angle
        if ((A!=0)&(A!=angle)){
            angle=A;
            Tourner(A*1.0, 0.4);
        }

        // On affiche la distance ou l'angle de commande sur Teraterm
        debug_pc.printf("\tCommande = ");
        debug_pc.printf("D=%d \t",D);
        debug_pc.printf("A=%d \t",A);
        debug_pc.printf("\r\n");
    }   
}